import textwrap

import aesara
import aesara.tensor as at
import pytest

from aeppl.printing import latex_pprint, pprint


def test_PreamblPPrinter():
    """Make sure we can print a `Function` and `FunctionGraph`."""
    srng = at.random.RandomStream(seed=2320)

    mu = at.scalar("\\mu")
    sigma = at.scalar("\\sigma")
    b = at.scalar("b")

    y = b * srng.normal(mu, sigma)

    y_fn = aesara.function([mu, sigma, b], y)

    expected = textwrap.dedent(
        r"""
    b in R, \mu in R, \sigma in R
    a ~ N(\mu, \sigma**2) in R
    (b * a)
    """
    )
    assert pprint(y_fn).strip() == expected.strip()


def test_notex_print():
    srng = at.random.RandomStream(seed=2320)

    normalrv_noname_expr = at.scalar("b") * srng.normal(
        at.scalar("\\mu"), at.scalar("\\sigma")
    )
    expected = textwrap.dedent(
        r"""
    b in R, \mu in R, \sigma in R
    a ~ N(\mu, \sigma**2) in R
    (b * a)
    """
    )
    assert pprint(normalrv_noname_expr) == expected.strip()

    # Make sure the constant shape is show in values and not symbols.
    normalrv_name_expr = at.scalar("b") * srng.normal(
        at.scalar("\\mu"), at.scalar("\\sigma"), size=[2, 1], name="X"
    )
    expected = textwrap.dedent(
        r"""
    b in R, \mu in R, \sigma in R
    X ~ N(\mu, \sigma**2) in R**(2 x 1)
    (b * X)
    """
    )
    assert pprint(normalrv_name_expr) == expected.strip()

    normalrv_noname_expr_2 = at.matrix("M") * srng.normal(
        at.scalar("\\mu_2"), at.scalar("\\sigma_2")
    )
    normalrv_noname_expr_2 *= at.scalar("b") * srng.normal(
        normalrv_noname_expr_2, at.scalar("\\sigma")
    ) + at.scalar("c")
    expected = textwrap.dedent(
        r"""
    M in R**(N^M_0 x N^M_1), \mu_2 in R, \sigma_2 in R
    b in R, \sigma in R, c in R
    a ~ N(\mu_2, \sigma_2**2) in R, d ~ N((M * a), \sigma**2) in R**(N^d_0 x N^d_1)
    ((M * a) * ((b * d) + c))
    """
    )
    assert pprint(normalrv_noname_expr_2) == expected.strip()

    expected = textwrap.dedent(
        r"""
    b in Z, c in Z, M in R**(N^M_0 x N^M_1)
    M[b, c]
    """
    )
    # TODO: "c" should be "1".
    assert (
        pprint(at.matrix("M")[at.iscalar("a"), at.constant(1, dtype="int")])
        == expected.strip()
    )

    expected = textwrap.dedent(
        r"""
    M in R**(N^M_0 x N^M_1)
    M[1]
    """
    )
    assert pprint(at.matrix("M")[1]) == expected.strip()

    expected = textwrap.dedent(
        r"""
    M in N**(N^M_0)
    M[2:4:0]
    """
    )
    assert pprint(at.vector("M", dtype="uint32")[0:4:2]) == expected.strip()


def test_tex_print():
    srng = at.random.RandomStream(seed=2320)

    normalrv_noname_expr = at.scalar("b") * srng.normal(
        at.scalar("\\mu"), at.scalar("\\sigma")
    )
    expected = textwrap.dedent(
        r"""
    \begin{equation}
      \begin{gathered}
        b \in \mathbb{R}, \,\mu \in \mathbb{R}, \,\sigma \in \mathbb{R}
        \\
        a \sim \operatorname{N}\left(\mu, {\sigma}^{2}\right)\,  \in \mathbb{R}
        \\
        (b \odot a)
      \end{gathered}
    \end{equation}
    """
    )
    assert latex_pprint(normalrv_noname_expr) == expected.strip()

    normalrv_name_expr = at.scalar("b") * srng.normal(
        at.scalar("\\mu"), at.scalar("\\sigma"), size=[2, 1], name="X"
    )
    expected = textwrap.dedent(
        r"""
    \begin{equation}
      \begin{gathered}
        b \in \mathbb{R}, \,\mu \in \mathbb{R}, \,\sigma \in \mathbb{R}
        \\
        X \sim \operatorname{N}\left(\mu, {\sigma}^{2}\right)\,  \in \mathbb{R}^{2 \times 1}
        \\
        (b \odot X)
      \end{gathered}
    \end{equation}
    """
    )
    assert latex_pprint(normalrv_name_expr) == expected.strip()

    normalrv_noname_expr_2 = at.matrix("M") * srng.normal(
        at.scalar("\\mu_2"), at.scalar("\\sigma_2")
    )
    normalrv_noname_expr_2 *= at.scalar("b") * srng.normal(
        normalrv_noname_expr_2, at.scalar("\\sigma")
    ) + at.scalar("c")
    expected = textwrap.dedent(
        r"""
    \begin{equation}
      \begin{gathered}
        M \in \mathbb{R}^{N^{M}_{0} \times N^{M}_{1}}
        \\
        \mu_2 \in \mathbb{R}, \,\sigma_2 \in \mathbb{R}
        \\
        b \in \mathbb{R}, \,\sigma \in \mathbb{R}, \,c \in \mathbb{R}
        \\
        a \sim \operatorname{N}\left(\mu_2, {\sigma_2}^{2}\right)\,  \in \mathbb{R}
        \\
        d \sim \operatorname{N}\left((M \odot a), {\sigma}^{2}\right)\,  \in \mathbb{R}^{N^{d}_{0} \times N^{d}_{1}}
        \\
        ((M \odot a) \odot ((b \odot d) + c))
      \end{gathered}
    \end{equation}
    """
    )
    assert latex_pprint(normalrv_noname_expr_2) == expected.strip()

    expected = textwrap.dedent(
        r"""
    \begin{equation}
      \begin{gathered}
        b \in \mathbb{Z}, \,c \in \mathbb{Z}, \,M \in \mathbb{R}^{N^{M}_{0} \times N^{M}_{1}}
        \\
        M\left[b, \,c\right]
      \end{gathered}
    \end{equation}
    """
    )
    # TODO: "c" should be "1".
    assert (
        latex_pprint(at.matrix("M")[at.iscalar("a"), at.constant(1, dtype="int")])
        == expected.strip()
    )

    expected = textwrap.dedent(
        r"""
    \begin{equation}
      \begin{gathered}
        M \in \mathbb{R}^{N^{M}_{0} \times N^{M}_{1}}
        \\
        M\left[1\right]
      \end{gathered}
    \end{equation}
    """
    )
    assert latex_pprint(at.matrix("M")[1]) == expected.strip()

    expected = textwrap.dedent(
        r"""
    \begin{equation}
      \begin{gathered}
        M \in \mathbb{N}^{N^{M}_{0}}
        \\
        M\left[2:4:0\right]
      \end{gathered}
    \end{equation}
    """
    )
    assert latex_pprint(at.vector("M", dtype="uint32")[0:4:2]) == expected.strip()

    S_rv = srng.invgamma(0.5, 0.5, name="S")
    T_rv = srng.halfcauchy(1.0, name="T")
    Y_rv = srng.normal(T_rv, at.sqrt(S_rv), name="Y")
    expected = textwrap.dedent(
        r"""
    \begin{equation}
      \begin{gathered}
        T \sim \operatorname{C^{+}}\left(1, 1\right)\,  \in \mathbb{R}
        \\
        S \sim \operatorname{Gamma^{-1}}\left(0.5, 0.5\right)\,  \in \mathbb{R}
        \\
        Y \sim \operatorname{N}\left(T, {\sqrt{S}}^{2}\right)\,  \in \mathbb{R}
        \\
        Y
      \end{gathered}
    \end{equation}
    """
    )
    assert latex_pprint(Y_rv) == expected.strip()


@pytest.mark.xfail(
    reason=r"AePPL is not aware of the distributions' support and displays \mathbb{R} by default"
)
def test_tex_print_support_dimension():
    srng = at.random.RandomStream(seed=2320)

    U_rv = srng.uniform(0, 1, name="U")
    T_rv = srng.halfcauchy(U_rv, name="T")
    expected = textwrap.dedent(
        r"""
    \begin{equation}
      \begin{gathered}
        U \sim \operatorname{U}\left(1, 1\right)\,  \in \left[0, 1\right]
        \\
        T \sim \operatorname{C^{+}}\left(1, 1\right)\,  \in \mathbb{R}^{+}
        \\
        T
      \end{gathered}
    \end{equation}
    """
    )
    assert latex_pprint(T_rv) == expected.strip()



from typing import (
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

import aesara.tensor as at
import numpy as np
from aesara import config
from aesara.graph.basic import Constant, Variable, clone_get_equiv, graph_inputs, walk
from aesara.graph.fg import FunctionGraph
from aesara.graph.op import compute_test_value
from aesara.graph.rewriting.utils import rewrite_graph
from aesara.link.c.type import CType
from aesara.tensor.rewriting.basic import topo_constant_folding
from aesara.tensor.rewriting.shape import ShapeFeature
from aesara.tensor.sharedvar import SharedVariable
from aesara.tensor.subtensor import AdvancedIncSubtensor, AdvancedIncSubtensor1
from aesara.tensor.var import TensorVariable

from aeppl.abstract import MeasurableVariable

PotentialShapeType = Union[
    int,
    np.ndarray,
    Tuple[Union[int, Variable], ...],
    List[Union[int, Variable]],
    Variable,
]


def change_rv_size(
    rv_var: TensorVariable,
    new_size: PotentialShapeType,
    expand: Optional[bool] = False,
) -> TensorVariable:
    """Change or expand the size of a `RandomVariable`.

    Parameters
    ==========
    rv_var
        The `RandomVariable` output.
    new_size
        The new size.
    expand:
        Expand the existing size by `new_size`.

    """
    rv_node = rv_var.owner
    rng, size, dtype, *dist_params = rv_node.inputs
    name = rv_var.name
    tag = rv_var.tag

    if expand:
        if rv_node.op.ndim_supp == 0 and at.get_vector_length(size) == 0:
            size = rv_node.op._infer_shape(size, dist_params)
        new_size = tuple(at.atleast_1d(new_size)) + tuple(size)

    # Make sure the new size is a tensor. This helps to not unnecessarily pick
    # up a `Cast` in some cases
    new_size = at.as_tensor(new_size, ndim=1, dtype="int64")

    new_rv_node = rv_node.op.make_node(rng, new_size, dtype, *dist_params)
    rv_var = new_rv_node.outputs[-1]
    rv_var.name = name
    for k, v in tag.__dict__.items():
        rv_var.tag.__dict__.setdefault(k, v)

    if config.compute_test_value != "off":
        compute_test_value(new_rv_node)

    return rv_var


def extract_obs_data(x: TensorVariable) -> np.ndarray:
    """Extract data from observed symbolic variables.

    Raises
    ------
    TypeError

    """
    if isinstance(x, Constant):
        return x.data
    if isinstance(x, SharedVariable):
        return x.get_value()
    if x.owner and isinstance(
        x.owner.op, (AdvancedIncSubtensor, AdvancedIncSubtensor1)
    ):
        array_data = extract_obs_data(x.owner.inputs[0])
        mask_idx = tuple(extract_obs_data(i) for i in x.owner.inputs[2:])
        mask = np.zeros_like(array_data)
        mask[mask_idx] = 1
        return np.ma.MaskedArray(array_data, mask)

    raise TypeError(f"Data cannot be extracted from {x}")


def walk_model(
    graphs: Iterable[TensorVariable],
    walk_past_rvs: bool = False,
    stop_at_vars: Optional[Set[TensorVariable]] = None,
    expand_fn: Callable[[TensorVariable], List[TensorVariable]] = lambda var: [],
) -> Generator[TensorVariable, None, None]:
    """Walk model graphs and yield their nodes.

    By default, these walks will not go past ``MeasurableVariable`` nodes.

    Parameters
    ==========
    graphs
        The graphs to walk.
    walk_past_rvs
        If ``True``, the walk will not terminate at ``MeasurableVariable``s.
    stop_at_vars
        A list of variables at which the walk will terminate.
    expand_fn
        A function that returns the next variable(s) to be traversed.
    """
    if stop_at_vars is None:
        stop_at_vars = set()

    def expand(var: TensorVariable, stop_at_vars=stop_at_vars) -> List[TensorVariable]:
        new_vars = expand_fn(var)

        if (
            var.owner
            and (walk_past_rvs or not isinstance(var.owner.op, MeasurableVariable))
            and (var not in stop_at_vars)
        ):
            new_vars.extend(reversed(var.owner.inputs))

        return new_vars

    yield from walk(graphs, expand, False)


def replace_rvs_in_graphs(
    graphs: Iterable[TensorVariable],
    replacement_fn: Callable[
        [TensorVariable, Dict[TensorVariable, TensorVariable]],
        Dict[TensorVariable, TensorVariable],
    ],
    initial_replacements: Optional[Dict[TensorVariable, TensorVariable]] = None,
    **kwargs,
) -> Tuple[TensorVariable, Dict[TensorVariable, TensorVariable]]:
    """Replace random variables in graphs.

    This will *not* recompute test values.

    Parameters
    ==========
    graphs
        The graphs in which random variables are to be replaced.

    Returns
    =======
    A ``tuple`` containing the transformed graphs and a ``dict`` of the
    replacements that were made.
    """
    replacements = {}
    if initial_replacements:
        replacements.update(initial_replacements)

    def expand_replace(var: TensorVariable) -> List[TensorVariable]:
        new_nodes: List[TensorVariable] = []
        if var.owner and isinstance(var.owner.op, MeasurableVariable):
            new_nodes.extend(replacement_fn(var, replacements))
        return new_nodes

    for var in walk_model(graphs, expand_fn=expand_replace, **kwargs):
        pass

    if replacements:
        inputs = [i for i in graph_inputs(graphs) if not isinstance(i, Constant)]
        equiv = {k: k for k in replacements.keys()}
        equiv = clone_get_equiv(inputs, graphs, False, False, equiv)

        fg = FunctionGraph(
            [equiv[i] for i in inputs],
            [equiv[o] for o in graphs],
            clone=False,
        )

        fg.replace_all(
            replacements.items(), import_missing=True, reason="replace_rvs_in_graphs"
        )

        graphs = list(fg.outputs)

    return graphs, replacements


def rvs_to_value_vars(
    graphs: Iterable[TensorVariable],
    initial_replacements: Optional[Dict[TensorVariable, TensorVariable]] = None,
    **kwargs,
) -> Tuple[TensorVariable, Dict[TensorVariable, TensorVariable]]:
    """Replace random variables in graphs with their value variables.

    This will *not* recompute test values in the resulting graphs.

    Parameters
    ==========
    graphs
        The graphs in which to perform the replacements.
    initial_replacements
        A ``dict`` containing the initial replacements to be made.

    """

    def replace_fn(var, replacements):
        rv_value_var = replacements.get(var, None)
        if rv_value_var is not None:
            replacements[var] = rv_value_var
            # In case the value variable is itself a graph, we walk it for
            # potential replacements
            return [rv_value_var]
        return []

    return replace_rvs_in_graphs(graphs, replace_fn, initial_replacements, **kwargs)


def convert_indices(indices, entry):
    if indices and isinstance(entry, CType):
        rval = indices.pop(0)
        return rval
    elif isinstance(entry, slice):
        return slice(
            convert_indices(indices, entry.start),
            convert_indices(indices, entry.stop),
            convert_indices(indices, entry.step),
        )
    else:
        return entry


def indices_from_subtensor(idx_list, indices):
    """Compute a useable index tuple from the inputs of a ``*Subtensor**`` ``Op``."""
    return tuple(
        tuple(convert_indices(list(indices), idx) for idx in idx_list)
        if idx_list
        else indices
    )


def get_constant_value(x):
    """Use constant folding to get a constant value."""

    axis_fg = FunctionGraph(outputs=[x], features=[ShapeFeature()], clone=True)

    (folded_axis,) = rewrite_graph(
        axis_fg, custom_rewrite=topo_constant_folding
    ).outputs

    if not isinstance(folded_axis, Constant):
        raise ValueError("Could not obtain a constant from constant folding")

    return folded_axis.data



import aesara
import aesara.tensor as at
import numpy as np
import scipy.stats as st

from aeppl import conditional_logprob, joint_logprob
from aeppl.censoring import MeasurableClip
from aeppl.rewriting import construct_ir_fgraph
from tests.utils import assert_no_rvs


def test_scalar_clipped_mixture():
    srng = at.random.RandomStream(0)

    x = at.clip(srng.normal(loc=1), 0.5, 1.5)
    x.name = "x"
    y = srng.beta(1, 2, name="y")

    comps = at.stack([x, y])
    comps.name = "comps"
    idxs = srng.bernoulli(0.4, name="idxs")
    mix = comps[idxs]
    mix.name = "mix"

    logp, (idxs_vv, mix_vv) = joint_logprob(idxs, mix)

    logp_fn = aesara.function([idxs_vv, mix_vv], logp)
    assert logp_fn(0, 0.4) == -np.inf
    assert np.isclose(logp_fn(0, 0.5), st.norm.logcdf(0.5, 1) + np.log(0.6))
    assert np.isclose(logp_fn(0, 1.3), st.norm.logpdf(1.3, 1) + np.log(0.6))
    assert np.isclose(logp_fn(1, 0.4), st.beta.logpdf(0.4, 1, 2) + np.log(0.4))


def test_nested_scalar_mixtures():
    srng = at.random.RandomStream(0)

    x = srng.normal(loc=-50, name="x")
    y = srng.normal(loc=50, name="y")
    comps1 = at.stack([x, y])
    comps1.name = "comps1"
    idxs1 = srng.bernoulli(0.5, name="idxs1")
    mix1 = comps1[idxs1]
    mix1.name = "mix1"

    w = srng.normal(loc=-100, name="w")
    z = srng.normal(loc=100, name="z")
    comps2 = at.stack([w, z])
    comps2.name = "comps2"
    idxs2 = srng.bernoulli(0.5, name="idxs2")
    mix2 = comps2[idxs2]
    mix2.name = "mix2"

    comps12 = at.stack([mix1, mix2])
    comps12.name = "comps12"
    idxs12 = srng.bernoulli(0.5, name="idxs12")
    mix12 = comps12[idxs12]
    mix12.name = "mix12"

    logp, (idxs1_vv, idxs2_vv, idxs12_vv, mix12_vv) = joint_logprob(
        idxs1, idxs2, idxs12, mix12
    )
    logp_fn = aesara.function([idxs1_vv, idxs2_vv, idxs12_vv, mix12_vv], logp)

    expected_mu_logpdf = st.norm.logpdf(0) + np.log(0.5) * 3
    assert np.isclose(logp_fn(0, 0, 0, -50), expected_mu_logpdf)
    assert np.isclose(logp_fn(0, 1, 0, -50), expected_mu_logpdf)
    assert np.isclose(logp_fn(1, 0, 0, 50), expected_mu_logpdf)
    assert np.isclose(logp_fn(1, 1, 0, 50), expected_mu_logpdf)
    assert np.isclose(logp_fn(0, 0, 1, -100), expected_mu_logpdf)
    assert np.isclose(logp_fn(0, 1, 1, 100), expected_mu_logpdf)
    assert np.isclose(logp_fn(1, 0, 1, -100), expected_mu_logpdf)
    assert np.isclose(logp_fn(1, 1, 1, 100), expected_mu_logpdf)

    assert np.isclose(logp_fn(0, 0, 0, 50), st.norm.logpdf(100) + np.log(0.5) * 3)
    assert np.isclose(logp_fn(0, 0, 1, 50), st.norm.logpdf(150) + np.log(0.5) * 3)


def test_unvalued_ir_reversion():
    """Make sure that un-valued IR rewrites are reverted."""
    srng = at.random.RandomStream(0)

    x_rv = srng.normal(name="X")
    y_rv = at.clip(x_rv, 0, 1)
    y_rv.name = "Y"
    z_rv = srng.normal(y_rv, 1, name="Z")
    z_vv = z_rv.clone()
    z_vv.name = "z"

    # Only the `z_rv` is "valued", so `y_rv` doesn't need to be converted into
    # measurable IR.
    rv_values = {z_rv: z_vv}

    z_fgraph, new_rvs_to_values = construct_ir_fgraph(rv_values)

    assert not any(isinstance(node.op, MeasurableClip) for node in z_fgraph.apply_nodes)


def test_shifted_cumsum():
    srng = at.random.RandomStream(0)

    x = srng.normal(size=(5,), name="x")
    y = 5 + at.cumsum(x)
    y.name = "y"

    y_vv = y.clone()
    logp, (y_vv,) = joint_logprob(y)
    assert np.isclose(
        logp.eval({y_vv: np.arange(5) + 1 + 5}),
        st.norm.logpdf(1) * 5,
    )


def test_double_log_transform_rv():
    srng = at.random.RandomStream(0)

    base_rv = srng.normal(0, 1)
    y_rv = at.log(at.log(base_rv))
    y_rv.name = "y"

    logp, (y_vv,) = conditional_logprob(y_rv)
    logp_fn = aesara.function([y_vv], logp[y_rv])

    log_log_y_val = np.asarray(0.5)
    log_y_val = np.exp(log_log_y_val)
    y_val = np.exp(log_y_val)
    np.testing.assert_allclose(
        logp_fn(log_log_y_val),
        st.norm().logpdf(y_val) + log_y_val + log_log_y_val,
    )


def test_affine_transform_rv():
    srng = at.random.RandomStream(0)

    loc = at.scalar("loc")
    scale = at.vector("scale")
    rv_size = 3

    y_rv = loc + srng.normal(0, 1, size=rv_size, name="base_rv") * scale
    y_rv.name = "y"

    logp, (y_vv,) = conditional_logprob(y_rv)
    assert_no_rvs(logp[y_rv])
    logp_fn = aesara.function([loc, scale, y_vv], logp[y_rv])

    loc_test_val = 4.0
    scale_test_val = np.full(rv_size, 0.5)
    y_test_val = np.full(rv_size, 1.0)

    np.testing.assert_allclose(
        logp_fn(loc_test_val, scale_test_val, y_test_val),
        st.norm(loc_test_val, scale_test_val).logpdf(y_test_val),
    )


def test_affine_log_transform_rv():
    srng = at.random.RandomStream(0)

    a, b = at.scalars("a", "b")
    base_rv = srng.lognormal(0, 1, name="base_rv", size=(1, 2))
    y_rv = a + at.log(base_rv) * b
    y_rv.name = "y"

    y_vv = y_rv.clone()

    logps, (y_vv,) = conditional_logprob(y_rv)
    logp = logps[y_rv]
    logp_fn = aesara.function([a, b, y_vv], logp)

    a_val = -1.5
    b_val = 3.0
    y_val = [[0.1, 0.1]]

    assert np.allclose(
        logp_fn(a_val, b_val, y_val),
        st.norm(a_val, b_val).logpdf(y_val),
    )
